{-# LANGUAGE GeneralizedNewtypeDeriving #-}
{-# LANGUAGE DeriveGeneric #-}
{-# LANGUAGE FlexibleInstances #-}
{-# LANGUAGE FlexibleContexts #-}
{-# LANGUAGE MultiParamTypeClasses #-}
{-# LANGUAGE OverloadedStrings #-}
{-# LANGUAGE TypeFamilies #-}
{-# LANGUAGE PackageImports #-}

{-# LANGUAGE NoImplicitPrelude #-}

module Datarekisteri.Backend.Types where

import Relude

import Control.Monad.Except (throwError)
import Control.Monad.Logger (LoggingT)
import Datarekisteri.Core.Types
import Data.Aeson (ToJSON(..), FromJSON(..))
import Data.ByteArray (ByteArray, ByteArrayAccess)
import Data.Morpheus.App.Internal.Resolving (Resolver, LiftOperation)
import Data.Morpheus.Server.Types (SCALAR)
import Data.Morpheus.Types (GQLType, DecodeScalar(..), KIND, EncodeScalar(..),
    ScalarValue(..), MonadError, GQLError)
import Data.Morpheus.Types.GQLScalar (scalarToJSON, scalarFromJSON)
import Data.Time (UTCTime, getCurrentTime, NominalDiffTime, addUTCTime, Day)
import Data.Time.Format.ISO8601 (iso8601Show, iso8601ParseM)
import qualified Data.Text as T
import Data.Char (isSpace)
import Database.Persist.Class (PersistField(..))
import Database.Persist.PersistValue (PersistValue(..))
import Database.Persist.Sql (PersistFieldSql(..), SqlBackend)
import Network.Mail.Mime (Mail, Address(..))
import Text.Email.Validate (EmailAddress, toByteString, validate, emailAddress)
import qualified "base64" Data.ByteString.Base64 as B64 (encodeBase64, decodeBase64)
import "cryptonite" Crypto.Random (MonadRandom(..))

forward :: Monad m => [a] -> m [Maybe a]
forward = pure . map Just

requirePermission :: (MonadPermissions m, MonadError GQLError m) => Scope -> Permission -> m ()
requirePermission scope permission = unlessM (hasPermission scope permission) $
    throwError $ "Insufficient permissions, " <> show permission <> " for "
        <> show scope <> " required."

data MemberData = MemberData
    { name :: Text
    , nickname :: Maybe Text
    , birthdate :: Date
    , homeplace :: Text
    , application :: Text
    , phoneNumber :: PhoneNumber
    } deriving (Show, Eq, Generic)

instance FromJSON MemberData
instance ToJSON MemberData

newtype PasswordHash = PasswordHash ByteString
    deriving (Eq, Show, Ord, Semigroup, Monoid, ByteArrayAccess, ByteArray,
        PersistField, PersistFieldSql)

class Monad m => MonadTime m where
    currentTime :: m Time

instance MonadTime IO where
    currentTime = Time <$> getCurrentTime

class Monad m => MonadDB m where
    runQuery :: ReaderT SqlBackend (LoggingT IO) a -> m a
    -- TODO refactor so that it is possible to define a IO-less db, for safety and testability.
    --      Is there a way to do this that doesn't require moving all db calls to the class?
    --      Probably not :(

class Monad m => MonadEmail m where
    sendEmail :: Mail -> m ()
    fromAddress :: m Address

class Monad m => MonadRequest m where
    currentUser :: m (Maybe UserID)

class Monad m => MonadPermissions m where
    hasPermission :: Scope -> Permission -> m Bool
    currentPermissions :: m Text
    defaultPermissions :: m Text
    toPermissions :: Text -> m (Maybe Text)

instance (MonadDB m, LiftOperation o) => MonadDB (Resolver o () m) where
    runQuery = lift . runQuery

instance (MonadRequest m, LiftOperation o) => MonadRequest (Resolver o () m) where
    currentUser = lift currentUser

instance (MonadPermissions m, LiftOperation o) => MonadPermissions (Resolver o () m) where
    hasPermission scope permission = lift $ hasPermission scope permission
    defaultPermissions = lift defaultPermissions
    currentPermissions = lift currentPermissions
    toPermissions = lift . toPermissions

instance (MonadEmail m, LiftOperation o) => MonadEmail (Resolver o () m) where
    sendEmail = lift . sendEmail
    fromAddress = lift fromAddress

instance (MonadRandom m, LiftOperation o) => MonadRandom (Resolver o () m) where
    getRandomBytes = lift . getRandomBytes

instance (MonadTime m, LiftOperation o) => MonadTime (Resolver o () m) where
    currentTime = lift currentTime
